from typing import Annotated, Literal

from .base import Attrs as Attrs
from .base import NoAttrs as NoAttrs
from .styles import CSSProperties


class Alias(str):
    """Alias type for attributes."""


class HtmlAttrs(Attrs, total=False):
    """Common attributes for HTML elements."""

    id: str
    accesskey: str
    class_: Annotated[str, Alias("class")]
    classes: Annotated[list[str], Alias("class")]  # merged with class_
    contenteditable: Literal["true", "false"]
    dir: Literal["ltr", "rtl", "auto"]
    draggable: Literal["true", "false"]
    enterkeyhint: Literal["enter", "done", "go", "next", "previous", "search", "send"]
    hidden: Literal["true", "false"]
    inert: bool
    inputmode: Literal[
        "none", "text", "search", "tel", "url", "email", "numeric", "decimal"
    ]
    lang: str
    popover: bool
    spellcheck: Literal["true", "false"]
    style: CSSProperties
    tabindex: int
    title: str
    translate: Literal["yes", "no"]


class HtmxAttrs(Attrs, total=False):
    """HTMX attributes for HTML elements.

    See: https://htmx.org/
    """

    hx_get: Annotated[str, Alias("hx-get")]
    hx_post: Annotated[str, Alias("hx-post")]
    hx_put: Annotated[str, Alias("hx-put")]
    hx_delete: Annotated[str, Alias("hx-delete")]
    hx_patch: Annotated[str, Alias("hx-patch")]

    hx_on: Annotated[str, Alias("hx-on")]
    hx_include: Annotated[str, Alias("hx-include")]
    hx_confirm: Annotated[str, Alias("hx-confirm")]
    hx_trigger: Annotated[str, Alias("hx-trigger")]
    hx_target: Annotated[Literal["this", "next", "previous"] | str, Alias("hx-target")]
    hx_select: Annotated[str, Alias("hx-select")]
    hx_select_oob: Annotated[str, Alias("hx-select-oob")]
    hx_swap: Annotated[
        Literal[
            "innerHTML",
            "outerHTML",
            "beforebegin",
            "afterbegin",
            "beforeend",
            "afterend",
            "delete",
            "none",
        ]
        | str,
        Alias("hx-swap"),
    ]
    hx_swap_oob: Annotated[str, Alias("hx-swap-oob")]
    hx_vals: Annotated[str, Alias("hx-vals")]
    hx_sync: Annotated[str, Alias("hx-sync")]
    hx_boost: Annotated[bool, Alias("hx-boost")]
    hx_indicator: Annotated[str, Alias("hx-indicator")]
    hx_push_url: Annotated[bool, Alias("hx-push-url")]
    hx_history: Annotated[bool, Alias("hx-history")]
    hx_history_elt: Annotated[str, Alias("hx-history-elt")]
    hx_ext: Annotated[str, Alias("hx-ext")]
    hx_disable: Annotated[bool, Alias("hx-disable")]
    hx_disabled_ext: Annotated[str, Alias("hx-disabled-ext")]
    hx_disinherit: Annotated[str, Alias("hx-disinherit")]
    hx_encoding: Annotated[str, Alias("hx-encoding")]
    hx_headers: Annotated[str, Alias("hx-headers")]
    hx_params: Annotated[str, Alias("hx-params")]
    hx_preserve: Annotated[bool, Alias("hx-preserve")]
    hx_prompt: Annotated[str, Alias("hx-prompt")]
    hx_replace_url: Annotated[str, Alias("hx-replace-url")]
    hx_request: Annotated[str, Alias("hx-request")]
    hx_validate: Annotated[bool, Alias("hx-validate")]
    hx_ws: Annotated[str, Alias("hx-ws")]
    hx_sse: Annotated[str, Alias("hx-sse")]


class WindowEventAttrs(Attrs, total=False):
    """Event Attributes for HTML elements."""

    on_afterprint: Annotated[str, Alias("onafterprint")]
    on_beforeprint: Annotated[str, Alias("onbeforeprint")]
    on_beforeunload: Annotated[str, Alias("onbeforeunload")]
    on_error: Annotated[str, Alias("onerror")]
    on_hashchange: Annotated[str, Alias("onhashchange")]
    on_load: Annotated[str, Alias("onload")]
    on_message: Annotated[str, Alias("onmessage")]
    on_offline: Annotated[str, Alias("onoffline")]
    on_online: Annotated[str, Alias("ononline")]
    on_pagehide: Annotated[str, Alias("onpagehide")]
    on_pageshow: Annotated[str, Alias("onpageshow")]
    on_popstate: Annotated[str, Alias("onpopstate")]
    on_resize: Annotated[str, Alias("onresize")]
    on_storage: Annotated[str, Alias("onstorage")]
    on_unhandledrejection: Annotated[str, Alias("onunhandledrejection")]
    on_unload: Annotated[str, Alias("onunload")]


class FormEventAttrs(Attrs, total=False):
    """Event Attributes for HTML Form elements."""

    on_blur: Annotated[str, Alias("onblur")]
    on_change: Annotated[str, Alias("onchange")]
    on_contextmenu: Annotated[str, Alias("oncontextmenu")]
    on_focus: Annotated[str, Alias("onfocus")]
    on_input: Annotated[str, Alias("oninput")]
    on_invalid: Annotated[str, Alias("oninvalid")]
    on_reset: Annotated[str, Alias("onreset")]
    on_search: Annotated[str, Alias("onsearch")]
    on_select: Annotated[str, Alias("onselect")]
    on_submit: Annotated[str, Alias("onsubmit")]


class KeyboardEventAttrs(Attrs, total=False):
    """Event Attributes for Keyboard events."""

    on_keydown: Annotated[str, Alias("onkeydown")]
    on_keypress: Annotated[str, Alias("onkeypress")]
    on_keyup: Annotated[str, Alias("onkeyup")]


class MouseEventAttrs(Attrs, total=False):
    """Event Attributes for Mouse events."""

    on_click: Annotated[str, Alias("onclick")]
    on_dblclick: Annotated[str, Alias("ondblclick")]
    on_mousedown: Annotated[str, Alias("onmousedown")]
    on_mousemove: Annotated[str, Alias("onmousemove")]
    on_mouseout: Annotated[str, Alias("onmouseout")]
    on_mouseover: Annotated[str, Alias("onmouseover")]
    on_mouseup: Annotated[str, Alias("onmouseup")]
    on_wheel: Annotated[str, Alias("onwheel")]


class DragEventAttrs(Attrs, total=False):
    """Event Attributes for Drag events."""

    on_drag: Annotated[str, Alias("ondrag")]
    on_dragend: Annotated[str, Alias("ondragend")]
    on_dragenter: Annotated[str, Alias("ondragenter")]
    on_dragleave: Annotated[str, Alias("ondragleave")]
    on_dragover: Annotated[str, Alias("ondragover")]
    on_dragstart: Annotated[str, Alias("ondragstart")]
    on_drop: Annotated[str, Alias("ondrop")]


class ClipboardEventAttrs(Attrs, total=False):
    """Event Attributes for Clipboard events."""

    on_copy: Annotated[str, Alias("oncopy")]
    on_cut: Annotated[str, Alias("oncut")]
    on_paste: Annotated[str, Alias("onpaste")]


class EventAttrs(
    WindowEventAttrs,
    FormEventAttrs,
    KeyboardEventAttrs,
    MouseEventAttrs,
    DragEventAttrs,
    ClipboardEventAttrs,
    total=False,
):
    """Event Attributes for HTML elements."""


class HtmlAndEventAttrs(HtmlAttrs, EventAttrs, total=False):
    """Common HTML and event attributes."""


class GlobalAttrs(HtmxAttrs, HtmlAndEventAttrs, total=False):
    """Global attributes for HTML elements."""


class HtmlTagAttrs(Attrs, total=False):
    xmlns: str


class MetaAttrs(HtmlAttrs, total=False):
    name: str
    content: str
    charset: str


class StyleAttrs(HtmlAndEventAttrs, total=False):
    media: str
    type: Literal["text/css"]


class ScriptAttrs(HtmlAttrs, total=False):
    async_: bool
    crossorigin: Literal["anonymous", "use-credentials"]
    defer: bool
    integrity: str
    nomodule: bool
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    src: str
    type: str


class HeadLinkAttrs(HtmlAttrs, total=False):
    rel: Literal["canonical", "alternate", "stylesheet"]
    crossorigin: Literal["anonymous", "use-credentials"]
    hreflang: str
    href: str
    integrity: str


class InputAttrs(GlobalAttrs, total=False):
    accept: str
    alt: str
    autcomplete: Literal["on", "off"]
    autofocus: bool
    checked: bool
    dirname: str
    disabled: bool
    form: str
    formaction: str
    formenctype: Literal[
        "application/x-www-form-urlencoded", "multipart/form-data", "text/plain"
    ]
    formmethod: Literal["get", "post"]
    formnovalidate: bool
    formtarget: str
    height: int
    list: str
    max: int
    maxlength: int
    min: int
    minlength: int
    multiple: bool
    name: str
    pattern: str
    placeholder: str
    popovertarget: str
    popovertargetaction: Literal["show", "hide", "toggle"]
    readonly: bool
    required: bool
    size: int
    src: str
    step: int
    type: Literal[
        "checkbox",
        "button",
        "color",
        "date",
        "datetime-local",
        "email",
        "file",
        "hidden",
        "image",
        "month",
        "number",
        "password",
        "radio",
        "range",
        "reset",
        "search",
        "submit",
        "tel",
        "text",
        "time",
        "url",
        "week",
    ]
    value: str
    width: int


class OutputAttrs(GlobalAttrs, total=False):
    for_: Annotated[str, "for"]
    form: str
    name: str


class OptionAttrs(GlobalAttrs, total=False):
    disabled: bool
    label: str
    selected: bool
    value: str


class OptgroupAttrs(GlobalAttrs, total=False):
    disabled: bool
    label: str


class SelectAttrs(GlobalAttrs, total=False):
    autofocus: bool
    disabled: bool
    form: str
    multiple: bool
    name: str
    required: bool
    size: int


class TextAreaAttrs(GlobalAttrs, total=False):
    autofocus: bool
    cols: int
    dirname: str
    disabled: bool
    form: str
    maxlength: int
    name: str
    placeholder: str
    readonly: bool
    required: bool
    rows: int
    wrap: Literal["hard", "soft"]


class FieldsetAttrs(GlobalAttrs, total=False):
    disabled: bool
    form: str
    name: str


class BlockquoteAttrs(GlobalAttrs, total=False):
    cite: str


class FormAttrs(GlobalAttrs, total=False):
    accept_charset: str
    action: str
    autocomplete: Literal["on", "off"]
    enctype: Literal[
        "application/x-www-form-urlencoded", "multipart/form-data", "text/plain"
    ]
    method: Literal["get", "post"]
    name: str
    novalidate: bool
    rel: Literal[
        "external",
        "help",
        "license",
        "next",
        "nofollow",
        "noopener",
        "noreferrer",
        "opener",
        "prev",
        "search",
    ]
    target: str


class HyperlinkAttrs(GlobalAttrs, total=False):
    download: str
    href: str
    hreflang: str
    media: str
    ping: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    rel: Literal[
        "alternate",
        "author",
        "bookmark",
        "external",
        "help",
        "license",
        "next",
        "nofollow",
        "noreferrer",
        "noopener",
        "prev",
        "search",
        "tag",
    ]
    target: str
    type: str


class ButtonAttrs(GlobalAttrs, total=False):
    autofocus: bool
    disabled: bool
    form: str
    formaction: str
    formenctype: Literal[
        "application/x-www-form-urlencoded", "multipart/form-data", "text/plain"
    ]
    formmethod: Literal["get", "post"]
    formnovalidate: bool
    formtarget: str
    popovertarget: str
    popovertargetaction: Literal["show", "hide", "toggle"]
    name: str
    type: Literal["submit", "reset", "button"]
    value: str


class LabelAttrs(GlobalAttrs, total=False):
    for_: Annotated[str, "for"]
    name: str


class TdAttrs(GlobalAttrs, total=False):
    colspan: int
    headers: str
    rowspan: int


class ThAttrs(GlobalAttrs, total=False):
    abbr: str
    colspan: int
    headers: str
    rowspan: int
    scope: Literal["col", "colgroup", "row", "rowgroup"]


class LiAttrs(GlobalAttrs, total=False):
    value: int


class ImgAttrs(GlobalAttrs, total=False):
    alt: str
    crossorigin: Literal["anonymous", "use-credentials"]
    height: int
    ismap: bool
    loading: Literal["eager", "lazy"]
    longdesc: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "unsafe-url",
    ]
    sizes: str
    src: str
    srcset: str
    usemap: str
    width: int


class SvgAttrs(GlobalAttrs, total=False):
    version: str
    xmlns: str
    height: int
    width: int
    view_box: Annotated[str, Alias("viewBox")]
    clip_path: Annotated[str, Alias("clip-path")]
    clip_rule: Annotated[str, Alias("clip-rule")]
    fill: str
    fill_rule: Annotated[str, Alias("fill-rule")]
    stroke: str
    stroke_dasharray: Annotated[str, Alias("stroke-dasharray")]
    stroke_dashoffset: Annotated[str, Alias("stroke-dashoffset")]
    stroke_linecap: Annotated[str, Alias("stroke-linecap")]
    stroke_linejoin: Annotated[str, Alias("stroke-linejoin")]
    stroke_miterlimit: Annotated[str, Alias("stroke-miterlimit")]
    stroke_opacity: Annotated[str, Alias("stroke-opacity")]
    stroke_width: Annotated[str, Alias("stroke-width")]
    transform: str


class CircleAttrs(SvgAttrs, total=False):
    cx: str
    cy: str
    r: str


class LineAttrs(SvgAttrs, total=False):
    x1: str
    x2: str
    y1: str
    y2: str


class PathAttrs(SvgAttrs, total=False):
    d: str


class PolylineAttrs(SvgAttrs, total=False):
    points: str


class IframeAttrs(GlobalAttrs, total=False):
    allow: str
    allowfullscreen: bool
    allowpaymentrequest: bool
    height: int
    frameborder: str
    loading: Literal["eager", "lazy"]
    name: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    sandbox: Literal[
        "allow-forms",
        "allow-pointer-lock",
        "allow-popups",
        "allow-same-origin",
        "allow-scripts",
        "allow-top-navigation",
    ]
    src: str
    srcdoc: str
    scrolling: str
    width: int


class ColAttrs(GlobalAttrs, total=False):
    span: int


class DelAttrs(GlobalAttrs, total=False):
    cite: str
    datetime: str


class InsAttrs(GlobalAttrs, total=False):
    cite: str
    datetime: str


class AreaAttrs(GlobalAttrs, total=False):
    alt: str
    coords: str
    download: str
    href: str
    ping: str
    referrerpolicy: Literal[
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ]
    rel: str
    shape: Literal["rect", "circle", "poly", "default"]
    target: str


class SourceAttrs(GlobalAttrs, total=False):
    media: str
    sizes: str
    src: str
    srcset: str
    type: str


class AudioAttrs(GlobalAttrs, total=False):
    src: str
    preload: Literal["auto", "metadata", "none"]
    autoplay: bool
    controls: bool
    loop: bool
    muted: bool
    mediagroup: str


class BaseAttrs(GlobalAttrs, total=False):
    href: str
    target: str


class CanvasAttrs(GlobalAttrs, total=False):
    width: int
    height: int


class DataAttrs(GlobalAttrs, total=False):
    value: str


class DetailsAttrs(GlobalAttrs, total=False):
    open: bool


class DialogAttrs(GlobalAttrs, total=False):
    open: bool


class EmbedAttrs(GlobalAttrs, total=False):
    height: int
    src: str
    type: str
    width: int


class MapAttrs(GlobalAttrs, total=False):
    name: str


class MeterAttrs(GlobalAttrs, total=False):
    form: str
    high: int
    low: int
    max: int
    min: int
    optimum: int
    value: int


class ObjectAttrs(GlobalAttrs, total=False):
    data: str
    form: str
    height: int
    name: str
    type: str
    typemustmatch: bool
    usemap: str
    width: int


class OlAttrs(GlobalAttrs, total=False):
    reversed: bool
    start: int
    type: Literal["1", "a", "A", "i", "I"]


class ParamAttrs(GlobalAttrs, total=False):
    name: str
    value: str


class ProgressAttrs(GlobalAttrs, total=False):
    max: int
    value: int


class QAttrs(HtmlAttrs, total=False):
    cite: str


class TimeAttrs(GlobalAttrs, total=False):
    datetime: str


class TrackAttrs(GlobalAttrs, total=False):
    default: bool
    kind: Literal["subtitles", "captions", "descriptions", "chapters", "metadata"]
    label: str
    src: str
    srclang: str


class VideoAttrs(GlobalAttrs, total=False):
    autoplay: bool
    controls: bool
    height: int
    loop: bool
    muted: bool
    poster: str
    preload: Literal["auto", "metadata", "none"]
    src: str
    width: int



import inspect
from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar, get_origin

from starlette._utils import is_async_callable
from starlette.concurrency import run_in_threadpool
from starlette.datastructures import FormData, Headers, QueryParams
from starlette.requests import Request
from starlette.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
    StreamingResponse,
)
from starlette.websockets import WebSocket

from ludic.types import BaseElement
from ludic.web import datastructures as ds
from ludic.web.parsers import BaseParser

__all__ = (
    "LudicResponse",
    "Response",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "FileResponse",
    "StreamingResponse",
    "RedirectResponse",
)

T = TypeVar("T")
P = ParamSpec("P")


async def run_in_threadpool_safe(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    def func_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        with BaseElement.formatter:
            return func(*args, **kwargs)

    response: T = await run_in_threadpool(func_wrapped, *args, **kwargs)
    return response


async def prepare_response(
    handler: Callable[..., Any],
    request: Request,
    status_code: int | None = None,
    headers: Headers | None = None,
) -> Response:
    """Prepares response for the given handler and request.

    Args:
        handler: The handler to prepare the response for.
        request: The request to prepare the response for.
        status_code: The status code to use for the response.
        headers: The headers to use for the response.

    Returns:
        The prepared response.
    """
    handler_kw = await extract_from_request(handler, request)
    is_async = is_async_callable(handler)

    if is_async:
        with BaseElement.formatter:
            raw_response = await handler(**handler_kw)
    else:
        raw_response = await run_in_threadpool_safe(handler, **handler_kw)

    if isinstance(raw_response, tuple):
        if len(raw_response) == 2:
            raw_response, status_or_headers = raw_response
            if isinstance(status_or_headers, dict):
                headers = ds.Headers(status_or_headers)
            else:
                status_code = status_or_headers
        elif len(raw_response) == 3:
            raw_response, status_code, headers = raw_response
            headers = ds.Headers(headers)
        else:
            raise ValueError(f"Invalid response tuple: {raw_response}")

    if raw_response is None:
        raw_response = ""

    response: Response
    if isinstance(raw_response, BaseElement):
        raw_response.context["request"] = request
        response = LudicResponse(
            raw_response, status_code=status_code or 200, headers=headers
        )
    elif isinstance(raw_response, str | bool | int | float):
        response = PlainTextResponse(
            str(raw_response), status_code=status_code or 200, headers=headers
        )
    elif isinstance(raw_response, Response):
        response = raw_response
    else:
        raise ValueError(f"Invalid response type: {type(raw_response)}")

    return response


async def extract_from_request(
    handler: Callable[..., Any],
    request: Request | WebSocket,
) -> dict[str, Any]:
    """Extracts parameters for given handler from the request.

    This function scans the signature of the handler and tries to extract
    the parameters from the request. It passes them to the handler as
    keyword arguments.
    """
    parameters = inspect.signature(handler).parameters
    handler_kwargs: dict[str, Any] = {}

    if all(key in parameters for key in request.path_params):
        handler_kwargs.update(request.path_params)

    for name, param in parameters.items():
        if isinstance(request, WebSocket):
            if issubclass(param.annotation, WebSocket):
                handler_kwargs[name] = request
        elif (origin := get_origin(param.annotation)) is not None and issubclass(
            origin, BaseParser
        ):
            async with request.form() as form:
                handler_kwargs[name] = param.annotation(form)
        elif issubclass(param.annotation, FormData):
            async with request.form() as form:
                handler_kwargs[name] = form
        elif issubclass(param.annotation, Request):
            handler_kwargs[name] = request
        elif issubclass(param.annotation, QueryParams):
            handler_kwargs[name] = request.query_params
        elif issubclass(param.annotation, Headers):
            handler_kwargs[name] = request.headers

    return handler_kwargs


class LudicResponse(HTMLResponse):
    """Response class for Ludic components."""

    def render(self, content: BaseElement) -> bytes:
        return content.to_html().encode("utf-8")



from typing import Annotated, NotRequired, Self, override

from examples import Page, init_db

from ludic.attrs import Attrs, HtmxAttrs
from ludic.catalog.buttons import (
    ButtonPrimary,
    ButtonSecondary,
    ButtonSuccess,
)
from ludic.catalog.forms import InputField
from ludic.catalog.headers import H1, H2
from ludic.catalog.layouts import Cluster
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import ColumnMeta, Table, TableHead, TableRow
from ludic.types import JavaScript
from ludic.web import Endpoint, LudicApp
from ludic.web.exceptions import NotFoundError
from ludic.web.parsers import Parser

db = init_db()
app = LudicApp(debug=True)


class PersonAttrs(Attrs):
    id: NotRequired[str]
    name: Annotated[str, ColumnMeta()]
    email: Annotated[str, ColumnMeta()]


class PeopleAttrs(Attrs):
    people: list[PersonAttrs]


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Edit Row"),
        Quote(
            "This example shows how to implement editable rows.",
            source_url="https://htmx.org/examples/edit-row/",
        ),
        H2("Demo"),
        await PeopleTable.get(),
    )


@app.endpoint("/people/{id}")
class PersonRow(Endpoint[PersonAttrs]):
    on_click_script: JavaScript = JavaScript(
        """
        let editing = document.querySelector('.editing')

        if (editing) {
            alert('You are already editing a row')
        } else {
            htmx.trigger(this, 'edit')
        }
        """
    )

    @classmethod
    async def put(cls, id: str, data: Parser[PersonAttrs]) -> Self:
        person = db.people.get(id)

        if person is None:
            raise NotFoundError("Person not found")

        for attr, value in data.validate().items():
            setattr(person, attr, value)

        return cls(**person.dict())

    @classmethod
    async def get(cls, id: str) -> Self:
        person = db.people.get(id)

        if person is None:
            raise NotFoundError("Person not found")

        return cls(**person.dict())

    @override
    def render(self) -> TableRow:
        return TableRow(
            self.attrs["name"],
            self.attrs["email"],
            ButtonPrimary(
                "Edit",
                hx_get=self.url_for(PersonForm),
                hx_trigger="edit",
                on_click=self.on_click_script,
                classes=["small"],
            ),
        )


@app.endpoint("/people/{id}/form/")
class PersonForm(Endpoint[PersonAttrs]):
    @classmethod
    async def get(cls, id: str) -> Self:
        person = db.people.get(id)

        if person is None:
            raise NotFoundError("Person not found")

        return cls(**person.dict())

    @override
    def render(self) -> TableRow:
        return TableRow(
            InputField(name="name", value=self.attrs["name"]),
            InputField(name="email", value=self.attrs["email"]),
            Cluster(
                ButtonSecondary(
                    "Cancel",
                    hx_get=self.url_for(PersonRow),
                    classes=["small"],
                ),
                ButtonSuccess(
                    "Save",
                    hx_put=self.url_for(PersonRow),
                    hx_include="closest tr",
                    classes=["small"],
                ),
                classes=["cluster-small"],
            ),
            classes=["editing"],
        )


@app.endpoint("/people/")
class PeopleTable(Endpoint[PeopleAttrs]):
    @classmethod
    async def get(cls) -> Self:
        return cls(people=[person.dict() for person in db.people.values()])

    @override
    def render(self) -> Table[TableHead, PersonRow]:
        return Table[TableHead, PersonRow](
            TableHead("Name", "Email", "Action"),
            *(PersonRow(**person) for person in self.attrs["people"]),
            body_attrs=HtmxAttrs(hx_target="closest tr", hx_swap="outerHTML"),
            classes=["text-align-center"],
        )



from dataclasses import asdict, dataclass
from typing import Any, override

from ludic.attrs import NoAttrs
from ludic.catalog.layouts import Center, Stack
from ludic.catalog.pages import Body, Head, HtmlPage
from ludic.html import meta
from ludic.types import (
    AnyChildren,
    Component,
)


@dataclass
class Model:
    def dict(self) -> dict[str, Any]:
        return asdict(self)


@dataclass
class ContactData(Model):
    id: str
    first_name: str
    last_name: str
    email: str


@dataclass
class PersonData(Model):
    id: str
    name: str
    email: str
    active: bool = True


@dataclass
class DB:
    contacts: dict[str, ContactData]
    people: dict[str, PersonData]


def init_db() -> DB:
    return DB(
        contacts={
            "1": ContactData(
                id="1",
                first_name="John",
                last_name="Doe",
                email="qN6Z8@example.com",
            )
        },
        people={
            "1": PersonData(
                id="1",
                name="Joe Smith",
                email="joe@smith.org",
                active=True,
            ),
            "2": PersonData(
                id="2",
                name="Angie MacDowell",
                email="angie@macdowell.org",
                active=True,
            ),
            "3": PersonData(
                id="3",
                name="Fuqua Tarkenton",
                email="fuqua@tarkenton.org",
                active=True,
            ),
            "4": PersonData(
                id="4",
                name="Kim Yee",
                email="kim@yee.org",
                active=False,
            ),
        },
    )


class Page(Component[AnyChildren, NoAttrs]):
    @override
    def render(self) -> HtmlPage:
        return HtmlPage(
            Head(
                meta(charset="utf-8"),
                meta(name="viewport", content="width=device-width, initial-scale=1.0"),
                title="HTMX Examples",
            ),
            Body(
                Center(
                    Stack(*self.children, id="content"),
                    style={"padding": self.theme.sizes.xxl},
                ),
                htmx_version="1.9.10",
            ),
        )



from ludic.catalog.forms import Form, InputField, TextAreaField
from ludic.catalog.headers import H1, H2, H3, H4, Anchor
from ludic.catalog.items import Key, Pairs, Value
from ludic.catalog.messages import (
    Message,
    MessageDanger,
    MessageInfo,
    MessageSuccess,
    MessageWarning,
    Title,
)
from ludic.catalog.navigation import Navigation, NavItem
from ludic.catalog.tables import Table, TableHead, TableRow
from ludic.catalog.typography import Link, Paragraph
from ludic.html import b
from ludic.styles import themes


def test_link() -> None:
    link = Link("A link!", to="https://example.com")
    assert link.to_html() == '<a href="https://example.com" target="_blank">A link!</a>'
    link = Link("A link!", to="/home")
    assert link.to_html() == '<a href="/home">A link!</a>'


def test_paragraph() -> None:
    paragraph = Paragraph(
        "Hello, how ",
        b("are you"),
        "? Click ",
        Link("here", to="https://example.com"),
        ".",
    )
    assert isinstance(paragraph.children[3], Link)
    assert paragraph.children[3].attrs["to"] == "https://example.com"
    assert paragraph.to_string(pretty=False) == (
        "<Paragraph>"
          'Hello, how <b>are you</b>? Click <Link to="https://example.com">here</Link>.'
        "</Paragraph>"
    )  # fmt: skip
    assert paragraph.to_html() == (
        '<p>'
            'Hello, how <b>are you</b>? '
            'Click <a href="https://example.com" target="_blank">here</a>.'
        '</p>'
    )  # fmt: skip


def test_navigation() -> None:
    navigation = Navigation(
        NavItem("Home", to="/"),
        NavItem("About", to="/about"),
        id="nav",
    )
    assert navigation.to_html() == (
        '<nav id="nav" class="navigation">'
            '<ul class="stack small">'
                '<li class="nav-item"><a href="/" class="btn">Home</a></li>'
                '<li class="nav-item">'
                    '<a href="/about" class="btn">About</a>'
                '</li>'
            "</ul>"
        "</nav>"
    )  # fmt: skip


def test_pairs() -> None:
    pairs = Pairs(
        Key("Name"),
        Value("John"),
        Key("Age"),
        Value(42),
    )
    assert pairs.to_html() == (
        '<dl>'
            '<dt>Name</dt>'
            '<dd>John</dd>'
            '<dt>Age</dt>'
            '<dd>42</dd>'
        "</dl>"
    )  # fmt: skip


def test_tables() -> None:
    table = Table(
        TableHead("Name", "Age"),
        TableRow("John", 42),
        TableRow("Jane", 43),
    )

    assert table.header == ("Name", "Age")
    assert table.getlist("Name") == ["John", "Jane"]
    assert table.getlist("Age") == [42, 43]
    assert table.getlist("NOT_FOUND") == []

    assert len(table.children) > 2
    assert table.children[2].get_value(123) is None

    assert table.to_html() == (
        '<table class="table">'
            "<thead>"
                '<tr><th>Name</th><th>Age</th></tr>'
            "</thead>"
            "<tbody>"
                '<tr><td>John</td><td>42</td></tr>'
                '<tr><td>Jane</td><td>43</td></tr>'
            "</tbody>"
        "</table>"
    )  # fmt: skip


def test_form_fields() -> None:
    form = Form(
        InputField(value="Name", name="name"),
        TextAreaField("Description", name="description"),
    )

    assert form.to_html() == (
        '<form class="form stack">'
            '<div class="form-field">'
                '<input value="Name" name="name" id="name" />'
            "</div>"
            '<div class="form-field">'
                '<textarea name="description" id="description">Description</textarea>'
            "</div>"
        "</form>"
    )  # fmt: skip

    form = Form(
        InputField(value="Name", name="name", label="Foo"),
        TextAreaField("Description", name="description", label="Bar"),
    )

    assert form.to_html() == (
        '<form class="form stack">'
            '<div class="form-field">'
                '<label for="name">Foo</label>'
                '<input value="Name" name="name" id="name" />'
            "</div>"
            '<div class="form-field">'
                '<label for="description">Bar</label>'
                '<textarea name="description" id="description">Description</textarea>'
            "</div>"
        "</form>"
    )  # fmt: skip


def test_header_anchor() -> None:
    assert Anchor(target="test").to_html() == '<a href="#test" class="anchor">#</a>'
    assert (
        Anchor("url", target="test").to_html()
        == '<a href="#test" class="anchor">url</a>'
    )


def test_headers() -> None:
    assert H1("Header 1").to_html() == "<h1>Header 1</h1>"
    assert H2("Header 2").to_html() == "<h2>Header 2</h2>"
    assert H3("Header 3").to_html() == "<h3>Header 3</h3>"
    assert H4("Header 4").to_html() == "<h4>Header 4</h4>"

    theme = themes.LightTheme(
        headers=themes.Headers(
            h1=themes.Header(anchor=True),
            h2=themes.Header(anchor=True),
            h3=themes.Header(anchor=True),
            h4=themes.Header(anchor=True),
        )
    )
    assert theme.use(H1("Header 1")).to_html() == (
        '<div class="with-anchor">'
            '<h1 id="header-1">Header 1</h1>'
            '<a href="#header-1" class="anchor">#</a>'
        '</div>'
    )  # fmt: skip
    assert theme.use(H2("Header 2")).to_html() == (
        '<div class="with-anchor">'
            '<h2 id="header-2">Header 2</h2>'
            '<a href="#header-2" class="anchor">#</a>'
        '</div>'
    )  # fmt: skip
    assert theme.use(H3("Header 3")).to_html() == (
        '<div class="with-anchor">'
            '<h3 id="header-3">Header 3</h3>'
            '<a href="#header-3" class="anchor">#</a>'
        '</div>'
    )  # fmt: skip
    assert theme.use(H4("Header 4")).to_html() == (
        '<div class="with-anchor">'
            '<h4 id="header-4">Header 4</h4>'
            '<a href="#header-4" class="anchor">#</a>'
        '</div>'
    )  # fmt: skip


def test_messages() -> None:
    assert Message("test").to_html() == (
        '<div class="message"><div class="content">test</div></div>'
    )
    assert Message(Title("Title"), "Content message").to_html() == (
        '<div class="message">'
        '<div class="title">Title</div>'
        '<div class="content">Content message</div>'
        "</div>"
    )
    assert MessageSuccess("test").to_html() == (
        '<div class="message success"><div class="content">test</div></div>'
    )
    assert MessageInfo("test").to_html() == (
        '<div class="message info"><div class="content">test</div></div>'
    )
    assert MessageWarning("test").to_html() == (
        '<div class="message warning"><div class="content">test</div></div>'
    )
    assert MessageDanger("test").to_html() == (
        '<div class="message danger"><div class="content">test</div></div>'
    )



from typing import override

from ludic.attrs import GlobalAttrs
from ludic.html import div, style
from ludic.types import AnyChildren, Component


class Title(Component[AnyChildren, GlobalAttrs]):
    classes = ["title"]

    @override
    def render(self) -> div:
        return div(*self.children)


class Message(Component[AnyChildren, GlobalAttrs]):
    classes = ["message"]
    styles = style.use(
        lambda theme: {
            ".message": {
                "background-color": theme.colors.white,
                "border": (f"{theme.borders.thin} solid {theme.colors.primary}"),
                "border-radius": theme.rounding.less,
                "font-size": theme.fonts.size * 0.9,
            },
            ".message > .title": {
                "font-weight": "bold",
                "border-top-left-radius": theme.rounding.less,
                "border-top-right-radius": theme.rounding.less,
                "padding-inline": theme.sizes.l,
                "padding-block": theme.sizes.s,
                "background-color": theme.colors.primary.lighten(1),
            },
            ".message > .content": {
                "padding-inline": theme.sizes.l,
                "padding-block": theme.sizes.s,
            },
        }
    )

    @override
    def render(self) -> div:
        if self.children and isinstance(self.children[0], Title):
            return div(
                self.children[0],
                div(*self.children[1:], classes=["content"]),
            )
        else:
            return div(div(*self.children, classes=["content"]))


class MessageSuccess(Message):
    classes = ["message", "success"]
    styles = style.use(
        lambda theme: {
            ".message.success > .title": {
                "background-color": theme.colors.success.lighten(1),
            },
            ".message.success": {
                "border-color": theme.colors.success,
            },
        }
    )


class MessageInfo(Message):
    classes = ["message", "info"]
    styles = style.use(
        lambda theme: {
            ".message.info > .title": {
                "background-color": theme.colors.info.lighten(1),
            },
            ".message.info": {
                "border-color": theme.colors.info,
            },
        }
    )


class MessageWarning(Message):
    classes = ["message", "warning"]
    styles = style.use(
        lambda theme: {
            ".message.warning > .title": {
                "background-color": theme.colors.warning.lighten(1),
            },
            ".message.warning": {
                "border-color": theme.colors.warning,
            },
        }
    )


class MessageDanger(Message):
    classes = ["message", "danger"]
    styles = style.use(
        lambda theme: {
            ".message.danger > .title": {
                "background-color": theme.colors.danger.lighten(1),
            },
            ".message.danger": {
                "border-color": theme.colors.danger,
            },
        }
    )



from typing import override

from ludic.attrs import ButtonAttrs
from ludic.html import button, style
from ludic.types import ComponentStrict, PrimitiveChildren

from .typography import Link


class Button(ComponentStrict[PrimitiveChildren, ButtonAttrs]):
    """Simple component creating an HTML button.

    The component creates a button.
    """

    classes = ["btn"]
    styles = style.use(
        lambda theme: {
            ".btn": {
                "display": "inline-block",
                "text-decoration": "none",
                "background-color": theme.colors.light,
                "color": theme.colors.black,
                "padding": f"{theme.sizes.xxxxs * 0.8} {theme.sizes.s}",
                "border": "none",
                "border-radius": theme.rounding.normal,
                "cursor": "pointer",
                "font-size": theme.fonts.size,
                "transition": "0.1s filter linear, 0.1s -webkit-filter linear",
            },
            ":not(a).btn": {
                "background-color": theme.colors.light.darken(2),
            },
            (".btn:hover", ".btn:focus"): {
                "filter": "brightness(85%)",
                "text-decoration": "none",
            },
            ".btn.small": {
                "font-size": theme.fonts.size * 0.9,
                "padding": f"{theme.sizes.xxxxs * 0.6} {theme.sizes.xxs}",
            },
            ".btn.large": {
                "font-size": theme.fonts.size * 1.2,
                "padding": f"{theme.sizes.xxxxs} {theme.sizes.m}",
            },
            (".invert .btn", ".active .btn"): {
                "background-color": theme.colors.dark,
                "color": theme.colors.light,
                "border-color": theme.colors.dark,
            },
        }
    )

    @override
    def render(self) -> button:
        return button(self.children[0], **self.attrs)


class ButtonPrimary(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``primary`` class.
    """

    classes = ["btn", "primary"]
    styles = style.use(
        lambda theme: {
            ".btn.primary": {
                "color": theme.colors.primary.readable(),
                "background-color": theme.colors.primary,
            }
        }
    )


class ButtonSecondary(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``secondary`` class.
    """

    classes = ["btn", "secondary"]
    styles = style.use(
        lambda theme: {
            ".btn.secondary": {
                "color": theme.colors.secondary.readable(),
                "background-color": theme.colors.secondary,
            }
        }
    )


class ButtonLink(Link):
    """Simple component creating an HTML button.

    The component creates a button with the ``secondary`` class.
    """

    classes = ["btn"]


class ButtonSuccess(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``success`` class.
    """

    classes = ["btn", "success"]
    styles = style.use(
        lambda theme: {
            ".btn.success": {
                "color": theme.colors.success.readable(),
                "background-color": theme.colors.success,
            }
        }
    )


class ButtonDanger(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``danger`` class.
    """

    classes = ["btn", "danger"]
    styles = style.use(
        lambda theme: {
            ".btn.danger": {
                "color": theme.colors.danger.readable(),
                "background-color": theme.colors.danger,
            }
        }
    )


class ButtonWarning(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``warning`` class.
    """

    classes = ["btn", "warning"]
    styles = style.use(
        lambda theme: {
            ".btn.warning": {
                "color": theme.colors.warning.readable(),
                "background-color": theme.colors.warning,
            }
        }
    )


class ButtonInfo(Button):
    """Simple component creating an HTML button.

    The component creates a button with the ``info`` class.
    """

    classes = ["btn", "info"]
    styles = style.use(
        lambda theme: {
            ".btn.info": {
                "color": theme.colors.info.readable(),
                "background-color": theme.colors.info,
            }
        }
    )



from typing import override

from ludic.attrs import Attrs
from ludic.html import a, blockquote, div, footer, p, style
from ludic.types import BaseElement, ComponentStrict


class QuoteAttrs(Attrs, total=False):
    source_url: str
    source_text: str


class Quote(ComponentStrict[str, QuoteAttrs]):
    """Simple component rendering as the HTML ``blockquote`` element."""

    classes = ["quote"]
    styles = style.use(
        lambda theme: {
            ".quote": {
                "blockquote": {
                    "background-color": theme.colors.light,
                    "border-left": (
                        f"{theme.borders.thick} solid {theme.colors.light.darken(1)}"
                    ),
                    "margin-bottom": theme.sizes.xxs,
                    "padding": f"{theme.sizes.l} {theme.sizes.m}",
                },
                "blockquote p": {
                    "font-size": theme.fonts.size,
                },
                "footer": {
                    "font-size": theme.fonts.size * 0.9,
                    "color": theme.colors.dark.lighten(1),
                },
                "footer a": {
                    "text-decoration": "none",
                    "color": theme.colors.dark.darken(1),
                },
                "footer a:hover": {
                    "text-decoration": "underline",
                },
            }
        }
    )

    @override
    def render(self) -> div:
        children: list[BaseElement] = [
            blockquote(*map(p, self.children[0].split("\n\n")))
        ]
        if source_url := self.attrs.get("source_url"):
            source_text = self.attrs.get("source_text", "Source: ")
            children.append(footer(source_text, a(source_url, href=source_url)))
        return div(*children)



from collections.abc import Callable, Iterator
from typing import Self, Unpack, override

from .attrs import (
    AreaAttrs,
    AudioAttrs,
    BaseAttrs,
    BlockquoteAttrs,
    ButtonAttrs,
    CanvasAttrs,
    CircleAttrs,
    ColAttrs,
    DataAttrs,
    DelAttrs,
    DetailsAttrs,
    DialogAttrs,
    EmbedAttrs,
    FieldsetAttrs,
    FormAttrs,
    GlobalAttrs,
    HeadLinkAttrs,
    HtmlAttrs,
    HtmlTagAttrs,
    HyperlinkAttrs,
    IframeAttrs,
    ImgAttrs,
    InputAttrs,
    InsAttrs,
    LabelAttrs,
    LiAttrs,
    LineAttrs,
    MapAttrs,
    MetaAttrs,
    MeterAttrs,
    NoAttrs,
    ObjectAttrs,
    OlAttrs,
    OptgroupAttrs,
    OptionAttrs,
    OutputAttrs,
    ParamAttrs,
    PathAttrs,
    PolylineAttrs,
    ProgressAttrs,
    QAttrs,
    ScriptAttrs,
    SelectAttrs,
    SourceAttrs,
    StyleAttrs,
    SvgAttrs,
    TdAttrs,
    TextAreaAttrs,
    ThAttrs,
    TimeAttrs,
    TrackAttrs,
    VideoAttrs,
)
from .styles import (
    Theme,
    format_styles,
    from_components,
    from_loaded,
)
from .types import (
    AnyChildren,
    BaseElement,
    ComplexChildren,
    CSSProperties,
    Element,
    ElementStrict,
    GlobalStyles,
    NoChildren,
    PrimitiveChildren,
)


class div(Element[AnyChildren, GlobalAttrs]):
    html_name = "div"


class span(Element[AnyChildren, GlobalAttrs]):
    html_name = "span"


class main(Element[AnyChildren, GlobalAttrs]):
    html_name = "main"


class p(Element[AnyChildren, GlobalAttrs]):
    html_name = "p"


class a(Element[AnyChildren, HyperlinkAttrs]):
    html_name = "a"


class br(Element[NoChildren, GlobalAttrs]):
    html_name = "br"


class button(Element[AnyChildren, ButtonAttrs]):
    html_name = "button"


class label(Element[AnyChildren, LabelAttrs]):
    html_name = "label"


class td(Element[AnyChildren, TdAttrs]):
    html_name = "td"


class th(Element[AnyChildren, ThAttrs]):
    html_name = "th"


class tr(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tr"


class thead(Element[ComplexChildren, GlobalAttrs]):
    html_name = "thead"


class tbody(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tbody"


class tfoot(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tfoot"


class table(Element[ComplexChildren, GlobalAttrs]):
    html_name = "table"


class li(Element[AnyChildren, LiAttrs]):
    html_name = "li"


class ul(Element[ComplexChildren, GlobalAttrs]):
    html_name = "ul"


class ol(Element[ComplexChildren, OlAttrs]):
    html_name = "ol"


class dt(Element[AnyChildren, GlobalAttrs]):
    html_name = "dt"


class dd(Element[AnyChildren, GlobalAttrs]):
    html_name = "dd"


class dl(Element[ComplexChildren, GlobalAttrs]):
    html_name = "dl"


class section(Element[AnyChildren, GlobalAttrs]):
    html_name = "section"


class input(Element[NoChildren, InputAttrs]):
    html_name = "input"


class output(Element[NoChildren, OutputAttrs]):
    html_name = "output"


class legend(Element[PrimitiveChildren, GlobalAttrs]):
    html_name = "legend"


class option(Element[PrimitiveChildren, OptionAttrs]):
    html_name = "option"


class optgroup(Element[AnyChildren, OptgroupAttrs]):
    html_name = "optgroup"


class select(Element[AnyChildren, SelectAttrs]):
    html_name = "select"


class textarea(Element[PrimitiveChildren, TextAreaAttrs]):
    always_pair = True
    html_name = "textarea"


class fieldset(Element[AnyChildren, FieldsetAttrs]):
    html_name = "fieldset"


class form(Element[AnyChildren, FormAttrs]):
    html_name = "form"


class img(Element[NoChildren, ImgAttrs]):
    html_name = "img"


class svg(Element[AnyChildren, SvgAttrs]):
    html_name = "svg"


class circle(Element[AnyChildren, CircleAttrs]):
    html_name = "circle"


class line(Element[AnyChildren, LineAttrs]):
    html_name = "line"


class path(Element[AnyChildren, PathAttrs]):
    html_name = "path"


class polyline(Element[AnyChildren, PolylineAttrs]):
    html_name = "polyline"


class b(Element[AnyChildren, GlobalAttrs]):
    html_name = "b"


class i(Element[AnyChildren, GlobalAttrs]):
    html_name = "i"


class s(Element[AnyChildren, GlobalAttrs]):
    html_name = "s"


class u(Element[AnyChildren, GlobalAttrs]):
    html_name = "u"


class strong(Element[AnyChildren, GlobalAttrs]):
    html_name = "strong"


class em(Element[AnyChildren, GlobalAttrs]):
    html_name = "em"


class mark(Element[AnyChildren, GlobalAttrs]):
    html_name = "mark"


class del_(Element[AnyChildren, DelAttrs]):
    html_name = "del"


class ins(Element[AnyChildren, InsAttrs]):
    html_name = "ins"


class header(Element[AnyChildren, GlobalAttrs]):
    html_name = "header"


class big(Element[AnyChildren, GlobalAttrs]):
    html_name = "big"


class small(Element[AnyChildren, GlobalAttrs]):
    html_name = "small"


class code(Element[AnyChildren, GlobalAttrs]):
    html_name = "code"


class pre(Element[AnyChildren, GlobalAttrs]):
    html_name = "pre"


class cite(Element[AnyChildren, GlobalAttrs]):
    html_name = "cite"


class blockquote(Element[AnyChildren, BlockquoteAttrs]):
    html_name = "blockquote"


class abbr(Element[AnyChildren, GlobalAttrs]):
    html_name = "abbr"


class h1(Element[AnyChildren, GlobalAttrs]):
    html_name = "h1"


class h2(Element[AnyChildren, GlobalAttrs]):
    html_name = "h2"


class h3(Element[AnyChildren, GlobalAttrs]):
    html_name = "h3"


class h4(Element[AnyChildren, GlobalAttrs]):
    html_name = "h4"


class h5(Element[AnyChildren, GlobalAttrs]):
    html_name = "h5"


class h6(Element[AnyChildren, GlobalAttrs]):
    html_name = "h6"


class title(Element[PrimitiveChildren, NoAttrs]):
    html_name = "title"


class link(Element[PrimitiveChildren, HeadLinkAttrs]):
    html_name = "link"


class style(BaseElement, GlobalStyles):
    html_name = "style"

    children: tuple[GlobalStyles | Callable[[Theme], GlobalStyles] | str]
    attrs: StyleAttrs

    def __init__(
        self,
        styles: GlobalStyles | Callable[[Theme], GlobalStyles] | str,
        theme: Theme | None = None,
        **attrs: Unpack[StyleAttrs],
    ) -> None:
        super().__init__(styles, **attrs)

        if theme:
            self.context["theme"] = theme

    @classmethod
    def use(cls, styles: GlobalStyles | Callable[[Theme], GlobalStyles]) -> Self:
        return cls(styles)

    @classmethod
    def from_components(
        cls, *components: type[BaseElement], theme: Theme | None = None
    ) -> Self:
        return cls(from_components(*components, theme=theme), type="text/css")

    @classmethod
    def load(cls, cache: bool = False, theme: Theme | None = None) -> Self:
        return cls(from_loaded(cache=cache, theme=theme), type="text/css")

    def __getitem__(self, key: str | tuple[str, ...]) -> CSSProperties | GlobalStyles:
        return self.styles[key]

    def __iter__(self) -> Iterator[str | tuple[str, ...]]:
        return iter(self.styles.keys())

    def __len__(self) -> int:
        return len(self.styles)

    @property
    def styles(self) -> GlobalStyles:
        if isinstance(self.children[0], str):
            return {}
        elif callable(self.children[0]):
            return self.children[0](self.theme)
        else:
            return self.children[0]

    @styles.setter
    def styles(self, value: GlobalStyles) -> None:
        self.children = (value,)

    def to_html(self) -> str:
        dom: BaseElement = self
        while dom != (rendered_dom := dom.render()):
            dom = rendered_dom

        attributes = ""
        if formatted_attrs := dom._format_attributes():
            attributes = f" {formatted_attrs}"

        if isinstance(dom.children[0], str):
            css_styles = self.children[0]
        else:
            css_styles = format_styles(self.styles)

        return (
            f"<{dom.html_name}{attributes}>\n"
            f"{css_styles}\n"
            f"</{dom.html_name}>"
        )  # fmt: off

    @override
    def render(self) -> BaseElement:
        return self


class script(Element[PrimitiveChildren, ScriptAttrs]):
    html_name = "script"
    always_pair = True


class noscript(Element[AnyChildren, GlobalAttrs]):
    html_name = "noscript"
    always_pair = True


class meta(Element[PrimitiveChildren, MetaAttrs]):
    html_name = "meta"


class head(Element[AnyChildren, NoAttrs]):
    html_name = "head"


class body(Element[AnyChildren, GlobalAttrs]):
    html_name = "body"


class footer(Element[AnyChildren, GlobalAttrs]):
    html_name = "footer"


class html(ElementStrict[head, body, HtmlTagAttrs]):
    html_header = "<!doctype html>"
    html_name = "html"


class iframe(Element[NoChildren, IframeAttrs]):
    always_pair = True
    html_name = "iframe"


class article(Element[AnyChildren, GlobalAttrs]):
    html_name = "article"


class address(Element[AnyChildren, GlobalAttrs]):
    html_name = "address"


class caption(Element[AnyChildren, GlobalAttrs]):
    html_name = "caption"


class col(Element[NoChildren, ColAttrs]):
    html_name = "col"


class colgroup(Element[AnyChildren, GlobalAttrs]):
    html_name = "colgroup"


class area(Element[NoChildren, AreaAttrs]):
    html_name = "area"


class aside(Element[AnyChildren, GlobalAttrs]):
    html_name = "aside"


class source(Element[NoChildren, SourceAttrs]):
    html_name = "source"


class audio(Element[AnyChildren, AudioAttrs]):
    html_name = "audio"


class base(Element[NoChildren, BaseAttrs]):
    html_name = "base"


class bdi(Element[AnyChildren, GlobalAttrs]):
    html_name = "bdi"


class bdo(Element[AnyChildren, GlobalAttrs]):
    html_name = "bdo"


class canvas(Element[AnyChildren, CanvasAttrs]):
    html_name = "canvas"


class data(Element[AnyChildren, DataAttrs]):
    html_name = "data"


class datalist(Element[AnyChildren, GlobalAttrs]):
    html_name = "datalist"


class details(Element[AnyChildren, DetailsAttrs]):
    html_name = "details"


class dfn(Element[AnyChildren, GlobalAttrs]):
    html_name = "dfn"


class dialog(Element[AnyChildren, DialogAttrs]):
    html_name = "dialog"


class embed(Element[NoChildren, EmbedAttrs]):
    html_name = "embed"


class figcaption(Element[AnyChildren, GlobalAttrs]):
    html_name = "figcaption"


class figure(Element[AnyChildren, GlobalAttrs]):
    html_name = "figure"


class hrgroup(Element[AnyChildren, GlobalAttrs]):
    html_name = "hrgroup"


class hr(Element[NoChildren, GlobalAttrs]):
    html_name = "hr"


class kbd(Element[AnyChildren, GlobalAttrs]):
    html_name = "kbd"


class map(Element[AnyChildren, MapAttrs]):
    html_name = "map"


class menu(Element[AnyChildren, GlobalAttrs]):
    html_name = "menu"


class meter(Element[AnyChildren, MeterAttrs]):
    html_name = "meter"


class nav(Element[AnyChildren, GlobalAttrs]):
    html_name = "nav"


class object(Element[AnyChildren, ObjectAttrs]):
    html_name = "object"


class param(Element[NoChildren, ParamAttrs]):
    html_name = "param"


class picture(Element[AnyChildren, GlobalAttrs]):
    html_name = "picture"


class progress(Element[AnyChildren, ProgressAttrs]):
    html_name = "progress"


class q(Element[AnyChildren, QAttrs]):
    html_name = "q"


class rp(Element[AnyChildren, GlobalAttrs]):
    html_name = "rp"


class rt(Element[AnyChildren, GlobalAttrs]):
    html_name = "rt"


class ruby(Element[AnyChildren, GlobalAttrs]):
    html_name = "ruby"


class samp(Element[AnyChildren, GlobalAttrs]):
    html_name = "samp"


class search(Element[AnyChildren, GlobalAttrs]):
    html_name = "search"


class sub(Element[AnyChildren, GlobalAttrs]):
    html_name = "sub"


class summary(Element[AnyChildren, GlobalAttrs]):
    html_name = "summary"


class sup(Element[AnyChildren, GlobalAttrs]):
    html_name = "sup"


class template(Element[AnyChildren, HtmlAttrs]):
    html_name = "template"


class time(Element[AnyChildren, TimeAttrs]):
    html_name = "time"


class track(Element[NoChildren, TrackAttrs]):
    html_name = "track"


class var(Element[AnyChildren, GlobalAttrs]):
    html_name = "var"


class video(Element[AnyChildren, VideoAttrs]):
    html_name = "video"


class wbr(Element[NoChildren, GlobalAttrs]):
    html_name = "wbr"



from ludic.catalog.typography import Link, Paragraph
from ludic.format import FormatContext, format_attr_value
from ludic.html import b, div, i, p, strong
from ludic.types import BaseElement


def test_format_attr_value() -> None:
    assert format_attr_value("foo", "bar") == "bar"
    assert format_attr_value("foo", "bar", is_html=True) == "bar"

    assert format_attr_value("foo", 1) == "1"
    assert format_attr_value("foo", 1, is_html=True) == "1"

    assert format_attr_value("foo", True) == "true"
    assert format_attr_value("foo", True, is_html=True) == "foo"

    assert format_attr_value("foo", False) == "false"
    assert format_attr_value("foo", False, is_html=True) == ""

    assert format_attr_value("hx-confirm", True) == "true"
    assert format_attr_value("hx-confirm", True, is_html=True) == "true"

    assert format_attr_value("hx-confirm", False) == "false"
    assert format_attr_value("hx-confirm", False, is_html=True) == "false"

    assert (
        format_attr_value("style", {"color": "red", "background": "blue"})
        == "color:red;background:blue"
    )


def test_format_context() -> None:
    with FormatContext("test_context") as ctx:
        first = ctx.append("foo")
        second = ctx.append({"bar": "baz"})
        extracts = ctx.extract(f"test {first} {second}")

    assert extracts == ["test ", "foo", " ", {"bar": "baz"}]


def test_format_context_in_elements() -> None:
    context = BaseElement.formatter
    assert context.get() == {}

    with context:
        f"test {b("foo")} {i("bar")}"
        assert list(context.get().values()) == [b("foo"), i("bar")]

    assert context.get() == {}

    with context:
        text = f"test {b("baz")} {i("foo")}"
        assert div(text) == div("test ", b("baz"), " ", i("foo"))
        assert div(f"test {div(f"foo {b("nested")}, {i("nested2")}")}") == div(
            "test ",
            div("foo ", b("nested"), ", ", i("nested2")),
        )

    assert context.get() == {}


def test_component_with_f_string() -> None:
    paragraph = Paragraph(
        f"Hello, how {strong("are you")}? Click {Link("here", to="https://example.com")}.",
    )
    assert isinstance(paragraph.children[3], Link)
    assert paragraph.children[3].attrs["to"] == "https://example.com"
    assert paragraph.to_string(pretty=False) == (
        "<Paragraph>"
          'Hello, how <strong>are you</strong>? '
          'Click <Link to="https://example.com">here</Link>.'
        "</Paragraph>"
    )  # fmt: skip
    assert paragraph.to_html() == (
        "<p>"
        "Hello, how <strong>are you</strong>? Click "
        '<a href="https://example.com" target="_blank">here</a>.'
        "</p>"
    )


def test_escaping_works() -> None:
    link = '<a href="https://example.com">test</a>'
    dom = p(f"Hello, how <b>are you</b>? Click {link}.")
    assert dom.to_html() == (
        "<p>Hello, how &lt;b&gt;are you&lt;/b&gt;? "
        'Click &lt;a href="https://example.com"&gt;test&lt;/a&gt;.</p>'
    )


def test_quotes_not_escaped() -> None:
    dom = p("It's alive <3.")
    assert dom.to_html() == "<p>It's alive &lt;3.</p>"



"""An experimental module for creating HTML forms."""

from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import Any, Literal, get_type_hints, override

from ludic.attrs import (
    Attrs,
    FormAttrs,
    InputAttrs,
    OptionAttrs,
    SelectAttrs,
    TextAreaAttrs,
)
from ludic.html import div, form, input, label, option, select, style, textarea
from ludic.types import (
    BaseElement,
    ComplexChildren,
    Component,
    NoChildren,
    PrimitiveChildren,
    TAttrs,
    TChildren,
)
from ludic.utils import get_annotations_metadata_of_type

from .utils import attr_to_camel

DEFAULT_FIELD_PARSERS: Mapping[str, Callable[[Any], PrimitiveChildren]] = {
    "checkbox": lambda value: True if value == "on" else False,
}


class SelectMetaAttrs(SelectAttrs):
    options: list[OptionAttrs]


@dataclass
class FieldMeta:
    """Class to be used in attributes annotations to create form fields.

    Example:

        def parse_email(email: str) -> str:
            if len(email.split("@")) != 2:
                raise ValidationError("Invalid email")
            return email

        class CustomerAttrs(Attrs):
            id: str
            name: Annotated[
                str,
                FieldMeta(label="Email", parser=parse_email),
            ]
    """

    label: str | Literal["auto"] | None = "auto"
    kind: Literal["input", "textarea", "checkbox", "select"] = "input"
    type: Literal["text", "email", "password", "hidden"] = "text"
    attrs: InputAttrs | TextAreaAttrs | SelectMetaAttrs | None = None
    parser: Callable[[Any], PrimitiveChildren] | None = None

    def format(self, key: str, value: Any) -> BaseElement:
        attrs = {} if self.attrs is None else dict(self.attrs)
        attrs["name"] = key

        if self.label == "auto":
            attrs["label"] = attr_to_camel(key)
        elif self.label:
            attrs["label"] = self.label

        match self.kind:
            case "input":
                return InputField(value=value, type=self.type, **attrs)
            case "checkbox":
                return InputField(checked=value, type="checkbox", **attrs)
            case "select":
                options: list[Option] = []

                for option_dict in attrs.pop("options", []):  # type: ignore
                    text = option_dict.pop("label", attr_to_camel(option_dict["value"]))
                    options.append(Option(text, **option_dict))

                return SelectField(*options, **attrs)
            case "textarea":
                return TextAreaField(value=value, **attrs)

    def parse(self, value: Any) -> PrimitiveChildren:
        if self.parser:
            return self.parser(value)
        else:
            return DEFAULT_FIELD_PARSERS.get(self.kind, str)(value)

    def __call__(self, value: Any) -> PrimitiveChildren:
        return self.parse(value)


class FieldAttrs(Attrs, total=False):
    """Shared attributes between custom form fields."""

    label: str


class InputFieldAttrs(FieldAttrs, InputAttrs):
    """Attributes of the component ``InputField``.

    The attributes are subclassed from :class:`FieldAttrs` and :class:`InputAttrs`.
    """


class SelectFieldAttrs(FieldAttrs, SelectAttrs):
    """Attributes of the component ``SelectField``.

    The attributes are subclassed from :class:`FieldAttrs` and :class:`SelectAttrs`.
    """


class TextAreaFieldAttrs(FieldAttrs, TextAreaAttrs):
    """Attributes of the component ``TextAreaField``.

    The attributes are subclassed from :class:`FieldAttrs` and :class:`TextAreaAttrs`.
    """


class FormField(Component[TChildren, TAttrs]):
    """Base class for form fields."""

    classes = ["form-field"]
    styles = style.use(
        lambda theme: {
            ".form-field label": {
                "display": "block",
                "font-weight": "bold",
                "margin-block-end": theme.sizes.xxs,
            },
            (".form-field input", ".form-field textarea", ".form-field select"): {
                "inline-size": "100%",
                "padding": f"{theme.sizes.xxxs} {theme.sizes.xs}",
                "border": (
                    f"{theme.borders.thin} solid {theme.colors.light.darken(2)}"
                ),
                "border-radius": theme.rounding.normal,
                "box-sizing": "border-box",
                "font-size": theme.fonts.size * 0.9,
                "transition": "all 0.3s ease-in-out",
                "resize": "vertical",
                "background-color": theme.colors.white,
            },
            (".form-field input", ".form-field select"): {
                "height": theme.sizes.xxxxl,
            },
            ".form-field input[type=checkbox]": {
                "height": theme.sizes.m,
                "width": theme.sizes.m,
            },
            (
                ".form-field input:focus",
                ".form-field textarea:focus",
                ".form-field select:focus",
            ): {
                "outline": "none",
                "border-color": theme.colors.light.darken(7),
                "border-width": theme.borders.thin,
            },
        }
    )

    def create_label(self, text: PrimitiveChildren, for_: str = "") -> label:
        if for_:
            return label(text, for_=for_)
        else:
            return label(text)


class InputField(FormField[NoChildren, InputFieldAttrs]):
    """Represents the HTML ``input`` element with an optional ``label`` element."""

    @override
    def render(self) -> div:
        attrs = self.attrs_for(input)
        if "name" in self.attrs:
            attrs["id"] = self.attrs["name"]

        elements: list[ComplexChildren] = []
        if text := self.attrs.get("label"):
            elements.append(self.create_label(text=text, for_=attrs.get("id", "")))
        elements.append(input(**attrs))

        return div(*elements)


class Option(Component[PrimitiveChildren, OptionAttrs]):
    """Represents the HTML ``option`` element."""

    @override
    def render(self) -> option:
        return option(*self.children, **self.attrs)


class SelectField(FormField[Option, SelectFieldAttrs]):
    """Represents the HTML ``input`` element with an optional ``label`` element."""

    @override
    def render(self) -> div:
        attrs = self.attrs_for(select)
        if "name" in self.attrs:
            attrs["id"] = self.attrs["name"]

        elements: list[ComplexChildren] = []
        if text := self.attrs.get("label"):
            elements.append(self.create_label(text=text, for_=attrs.get("id", "")))
        elements.append(select(*self.children, **attrs))

        return div(*elements)


class TextAreaField(FormField[PrimitiveChildren, TextAreaFieldAttrs]):
    """Represents the HTML ``textarea`` element with an optional ``label`` element."""

    @override
    def render(self) -> div:
        attrs = self.attrs_for(textarea)
        if "name" in self.attrs:
            attrs["id"] = self.attrs["name"]

        elements: list[ComplexChildren] = []
        if text := self.attrs.get("label"):
            elements.append(self.create_label(text=text, for_=attrs.get("id", "")))
        elements.append(textarea(*self.children, **attrs))

        return div(*elements)


class Form(Component[ComplexChildren, FormAttrs]):
    """A component helper for creating HTML forms."""

    classes = ["form", "stack"]

    @override
    def render(self) -> form:
        return form(*self.children, **self.attrs)


def create_fields(attrs: Any, spec: type[TAttrs]) -> tuple[ComplexChildren, ...]:
    """Create form fields from the given attributes.

    Example:

        class CustomerAttrs(Attrs):
            id: str
            name: Annotated[
                str,
                FieldMeta(label="Customer Name"),
            ]

        customer = CustomerAttrs(id=1, name="John Doe")
        fields = create_fields(customer, spec=CustomerAttrs)

        form = Form(*fields)

    Args:
        attrs (Any): The attributes to create form fields from.
        spec (type[TAttrs]): The specification of the attributes.

    Returns:
        ComplexChild: list of form fields.
    """
    annotations = get_type_hints(spec, include_extras=True)
    metadata_list = get_annotations_metadata_of_type(annotations, FieldMeta)
    fields: list[ComplexChildren] = []

    for name, metadata in metadata_list.items():
        if value := (attrs.get(name) or getattr(attrs, name, None)):
            field = metadata.format(name, value)
            fields.append(field)

    return tuple(fields)



"""A collection of Ludic components.

This module is meant as a collection of components that could be useful
for building Ludic applications. Any contributor is welcome to add new ones.

It also serves as showcase of possible implementations.
"""

# these need to be imported as they load important styling options
from . import pages, layouts  # noqa



import asyncio

from examples import Page

from ludic.catalog.headers import H1, H2, H3
from ludic.catalog.layouts import Box
from ludic.catalog.loaders import LazyLoader
from ludic.catalog.quotes import Quote
from ludic.web import LudicApp, Request

app = LudicApp(debug=True)


@app.get("/")
async def index(request: Request) -> Page:
    return Page(
        H1("Lazy Loading"),
        Quote(
            "This example shows how to lazily load an element on a page.",
            source_url="https://htmx.org/examples/lazy-load/",
        ),
        H2("Demo"),
        LazyLoader(load_url=request.url_for(load_svg, seconds=3)),
    )


@app.get("/load/{seconds:int}")
async def load_svg(seconds: int) -> Box:
    await asyncio.sleep(seconds)
    return Box(
        H3("Content Loaded!", classes=["text-align-center"]),
        classes=["invert"],
        style={"padding-top": "10rem", "padding-bottom": "10rem"},
    )



from .app import LudicApp
from .endpoints import Endpoint
from .requests import Request
from .responses import LudicResponse

__all__ = (
    "LudicApp",
    "Endpoint",
    "Request",
    "LudicResponse",
)
